#!/usr/bin/perl -w
#
# autoMT.pl v1.0
# Last modified March 2, 2006 11:55
# (c) Olaf R.P. Bininda-Emonds
#
# Purpose: Automates procedure of determinimg optimal model of evolution for DNA sequence data using
#		   either ModelTEST and PAUP* or MrAIC.pl and PHYML. Specifically, it can batch process all
#		   nexus- (*.nex) or phylip-foprmatted files (*.phylip), respectively in the current working
#		   directory and, for PAUP*-based analyses only, can also automatically test for the
#		   applicability of a global molecular clock using the LRT. for PAUP*-based analyses, it can
#		   also base the ModelTEST analysis on a user-input tree description in the data file rather
#		   than using the default NJ tree.
#
# Input: One or more nexus- (with extension .nex) or phylip-formatted (with extension .phylip) data files
#
# Output: A tab-delimited file containing ModelTEST / MrAIC.pl results for all input files
#
# Requires:
#	Any or all of the following in the $PATH: PAUP*, ModelTEST, and/or MrAIC.pl.
#	Path can also be specified directly in lines 45-47.
#
# Usage: autoMT.pl [-c] [-f<filename>] [-m<name>] [-p<name>] [-s<a|b|h>] [-v] [-h]
#	options: -a = use MrAIC.pl and PHYML to perform model fitting tests (default = PAUP* + ModelTEST command block)
#			 -c = do not test for the applicability of a global molecular clock (only applicable for PAUP*)
# 			 -f<filename> = name of a single file to be tested (default = autodetection and testing
#							of all *.nex or *.phylip files, as appropriate, in working directory)
#			 -m<name> = name of ModelTEST program in $PATH (default = modeltest)
#			 -p<name> = name of PAUP* program in $PATH (default = paup)
#			 -s<a|b|h> = use AIC (default), BIC (MrAIC only) or hLRT (PAUP* only) as selection criterion
#            -v = verbose output
#            -h = print this message and quit

use strict;
use File::Basename;
use constant PI => 3.1415926536;

# Set user-input defaults and associated parameters
	# I/O variables
		my @nexusFiles;
		my ($modelFitLog, $modelScores, $modelResults);
		
	# Program variables
		my $testProgram = "paup";	# Options are "paup" and "MrAIC"
		my $paup = "paup";
		my $modelTest = "modeltest";
		my $mraic = "/Users/bininda/bin/mraic.pl";
		
	# ModelTEST variables
		my $refTree = "NJ";
		my $defaultCriterion = "AIC";	# Options are "AIC" (default) and "hLRT"
		my $clockTest = 1;
		
	# Miscellaneous variables
		my $verbose = 0;
		my $debug = 0;
		my $version = "1.0";

# Process flags
	for (my $i = 0; $i <= $#ARGV; $i++)
		{
		if ($ARGV[$i] eq "-a")
			{
			$testProgram = "MrAIC";
			$clockTest = 0;
			}
		elsif ($ARGV[$i] eq "-c")
			{
			$clockTest = 0;
			}
		elsif ($ARGV[$i] =~ /^-f([\w+\.?\w+?]*)/)
			{
			die "ERROR: Nexus-formatted file $1 does not exist.\n" unless (-e $1);
			push @nexusFiles, $1;
			}
		elsif ($ARGV[$i] =~ /^-m(.+)/)
			{
			$modelTest = $1;
			}
		elsif ($ARGV[$i] =~ /^-p(.+)/)
			{
			$paup = $1;
			}
		elsif ($ARGV[$i] eq "-sa")
			{
			$defaultCriterion = "AIC";
			}
		elsif ($ARGV[$i] eq "-sb")
			{
			$defaultCriterion = "BIC";
			}
		elsif ($ARGV[$i] eq "-sh")
			{
			$defaultCriterion = "hLRT";
			}
		elsif ($ARGV[$i] eq "-v")
			{
			$verbose = 1;
			}
		elsif ($ARGV[$i] eq "-x")
			{
			$debug = $verbose = 1;
			}
		elsif ($ARGV[$i] eq "-h")
			{
			print "Usage: autoMT.pl [-c] [-f<filename>] [-m<name>] [-p<name>] [-s<a|b|h>] [-v] [-h]\n";
			print "Version: $version\n";
			print "Options: -a = use MrAIC.pl and PHYML to perform model fitting tests\n";
			print "              (default = PAUP* + ModelTEST command block)\n";
			print "         -c = do not test for the applicability of a global molecular clock (only applicable for PAUP*)\n";
			print "         -f<filename> = name of a single file to be tested (default = autodetection and testing\n";
			print "                        of all *.nex or *.phylip files, as appropriate, in working directory)\n";
			print "         -m<name> = name of ModelTEST program in \$PATH (default = modeltest)\n";
			print "         -p<name> = name of PAUP* program in \$PATH (default = paup)\n";
			print "         -s<a|b|h> = use AIC (default), BIC (MrAIC only) or hLRT (PAUP* only) as selection criterion\n";
			print "         -v = verbose output\n";
			print "         -h = print this message and quit\n";
			exit(0);
			}
		else
			{
			print "Don't understand argument: $ARGV[$i]\n";
			print "Usage: autoMT.pl [-c] [-f<filename>] [-m<name>] [-p<name>] [-s<a|h>] [-v] [-h]\n";
			print "Version: $version\n";
			exit(1); 
			}
		}
	
	$defaultCriterion = "AIC" if (($testProgram eq "MrAIC" and $defaultCriterion eq "hLRT") or ($testProgram eq "paup" and $defaultCriterion eq "BIC"));	# Correct for inapplicable criteria

# Get all nexus files in current directory if no file explicitly specified
	if (not @nexusFiles)
		{
		if ($testProgram eq "paup")
			{
			print "Getting all nexus-formatted files (*.nex) in current working directory ...\n\n";
			system("ls *.nex > nexusFiles.txt");
			}
		else
			{
			print "Getting all phylip-formatted files (*.phylip) in current working directory ...\n\n";
			system("ls *.phylip > nexusFiles.txt");
			}

		setLineBreak("nexusFiles.txt");
		open (NEX, "<nexusFiles.txt") or die "Cannot open file containing names of all nexus files, nexusFiles.txt\n";
			while (<NEX>)
				{
				chomp;
				next unless ($_);
				push @nexusFiles, $_ unless ($_ =~ /autoMT/i or $_ =~ /clockMT/i);	# Do not include autoMT temp files if present
				}
		close NEX;
		unlink ("nexusFiles.txt") unless ($debug);
		}
		
	die "ERROR: No files available to be tested.\n" unless (@nexusFiles);

# Open results file and print header row
	open (OUT, ">autoMTresults.txt") or die "Cannot write to results file autoMTresults.txt.\n";
		print OUT "Data file\tModel selected ($defaultCriterion)\tNonclock ln L";
			print OUT "\tClock ln L\tTaxon number\tdf\tChi-squared\tLRT P-value" if ($clockTest);
		print OUT "\tlsetcommand" if ($testProgram eq "paup");
		print OUT "\n";

# Run ModelTEST / MrAIC for all files
	foreach my $file (@nexusFiles)
		{
		my ($MTmodel, $nonclockL, $clockL, $ntax, $df, $chiSquared, $clockP, $LsetCommand);

		print "Processing file $file ...\n";
			unless (-e $file)
				{
				print "\tERROR: File does not exist; skipping ...\n";
				next;
				}
			my ($dataSource, $dir, $ext) = fileparse($file, qr/\.\D.*/);

		my $fileZero = time;

		if ($testProgram eq "paup")	# Run model-fitting using PAUP* and ModelTEST block
			{
			$modelFitLog = "modelfit_" . $dataSource . ".log";
			$modelScores = "model.scores_" . $dataSource;
			$modelResults = "MTresults_" . $dataSource . ".txt";
				
			# Check each file for tree description to be used using grep
				$refTree = "NJ";
				
				system ("grep -i tree\  $file > autoMTrefTree.txt");
				if (-s "autoMTrefTree.txt")
					{
					open (TREE, "<$file") or die "Cannot open file $file to check for any tree descriptions.\n";
						while (<TREE>)
							{
							last if ($refTree ne "NJ");	# Only store first tree description
							$refTree = $_ if ($_ =~ /\s*tree/i and $_ =~ /\(/ and $_ =~ /\)/ and $_ =~ /\;\s*$/)	# Basic tests to ensure that a tree description is being matched
							}
						close TREE;
					}
				
			# Create nexus file with ModelTEST instructions
				printMT("MTinstruct.txt");
				system ("cat $file MTinstruct.txt > autoMT.nex");
	
			# Analyze file and remove temp files
				print "\tRunning ModelTEST paup block in PAUP* ...\n";
				system ("$paup autoMT.nex > nonclockResults.txt");
				system ("$modelTest < $modelScores > $modelResults");
				my $MTtime = time - $fileZero;
					if ($MTtime > 3600)
						{
						printf "\t\tTime taken: %.1f hours", $MTtime / 3600;
						}
					else
						{
						print "\t\tTime taken: $MTtime seconds ...\n";
						}
				
			# Parse ModelTEST results file
				my ($MTcriterion, $lnL);
				open (RESULTS, "<$modelResults") or die "Cannot open file containing results of ModelTEST analysis, $modelResults\n";
					while (<RESULTS>)
						{
						chomp;
						next unless ($_);
						next unless ($_ =~ /^Lset/ or $_ =~ /^Likelihood settings/ or $_ =~ /^\s+\-lnL/);
						my $MTline = $_;
						
						if ($MTline =~ /^\s+\-lnL/)	# Get log-likelihood
							{
							(my $header, $lnL) = split(/=/, $MTline);
								$lnL =~ s/\s//g;
							}
	
						if ($MTline =~ /^Likelihood settings/)	# Get criterion and selected model
							{
							$MTcriterion = "hLRT" if ($MTline =~ /selected by hLRT/);
							$MTcriterion = "AIC" if ($MTline =~ /selected by AIC/);
							
							if ($MTline =~ /model \((\w+\+?\w?\+?\w?)\)/)
								{
								$MTmodel = $1 if ($MTcriterion eq $defaultCriterion);
								}
							}
	
						if ($MTline =~ /^Lset/)	# Get lset command for clock testing; store lnL
							{
							if ($MTcriterion eq $defaultCriterion)
								{
								$LsetCommand = $MTline;
									$LsetCommand =~ s/\s+$//;
								$nonclockL = $lnL;
								}
							}
						}
				close RESULTS;
			
			# Do clock-related tests
				if ($clockTest)
					{
					print "\tCalculating ln L for data under optimal model under assumption of a molecular clock ...\n";
					(my $clockCommand = $LsetCommand) =~ s/\;$/ clock=yes;/;
						$clockCommand =~ s/0\.0000/0\.0001/g;	# Remove impossible rate-matrix parameters
					my $clockZero = time;
					
					# Get data set characteristics using grep
						system ("grep -i -m 1 ntax $file > autoMTntax.txt");
						open (NTAX, "<autoMTntax.txt") or die "Cannot open text file autoMTntax.txt containing ntax line from $file.\n";
							while (<NTAX>)
								{
								if ($_ =~ /ntax\s*\=\s*(\d+)/)
									{
									$ntax = $1;
									}
								}
						close NTAX;
					
					# Create and analyze nexus file with ModelTEST instructions
						open (CLOCK, ">MTclock.txt") or die "Cannot write to clock instruction file MTclock.txt.\n";
							if ($refTree ne "NJ")
								{
								print CLOCK "\nbegin paup;\n\tset warnroot = no;\nend;\n";
								print CLOCK "\nbegin trees;\n\t$refTree\nend;\n";
								}
							
							print CLOCK "\nbegin paup;\n";
							print CLOCK "\tlog file = clockfit_" . $file . ".log replace;\n";
							print CLOCK "\n\tset criterion=likelihood;\n";
							print CLOCK "\n\t$clockCommand\n";
							if ($refTree eq "NJ")
								{
								print CLOCK "\n\thsearch start=nj swap=none;\n";
								}
							print CLOCK "\n\tdescribetrees / plot = phylogram outroot = monophyl brlens = no;\n";
							print CLOCK "\tlog stop;\n";
							print CLOCK "\nend;\n";
							print CLOCK "\nquit warntsave=no;\n";
						close CLOCK;
						
						system ("cat $file MTclock.txt > clockMT.nex");
						system ("$paup clockMT.nex > clockResults.txt");
						
						open (LNL, "<clockResults.txt") or die "Cannot open results from enforced clock analysis, clockResults.txt.\n";
							while (<LNL>)
								{
								chomp;
								last unless (not defined $clockL);
								next unless ($_ =~ /^\-Ln likelihood \=/);
								
								(my $header, $clockL) = split(/=/, $_);
									$clockL =~ s/\s//g;
								}
						close LNL;
	
						printf "\t\tTime taken: %s seconds\n", time - $clockZero;
						
					# Perform LRT to test applicability of clock
						$chiSquared =  sprintf ("%.5g", (2 * ($clockL - $nonclockL)));
						$df = $ntax - 2;
						
						$clockP = sprintf ("%.4g", _subchisqrprob($df, $chiSquared));
						
					# Remove temp files
						unless ($debug)
							{
							unlink("MTclock.txt", "clockMT.nex", "clockResults.txt");
							unlink ("MTinstruct.txt", "autoMT.nex", "nonclockResults.txt", "autoMTrefTree.txt", "autoMTntax.txt");
							}
					}
			}
		else	# Run model-fitting using PHYML and MrAIC.pl
			{
			# Run model-fitting
				system ("$mraic -modeltest $file > autoMT_phyml.txt");
				
			# Parse output file
				my $MTcriterion = $defaultCriterion;
				my $lnL;
				my %optimalModel;
				my $outputCount;

				$modelResults = $file . ".MrAIC.txt";
				open (RESULTS, "<$modelResults") or die "Cannot open file containing results of ModelTEST analysis, $modelResults\n";
					while (<RESULTS>)
						{
						chomp;
						next unless ($_);
						next if ($_ =~ /Nylander/ or $_ =~ /^Input data from/ or $_ =~ /^Note:/ or $_ =~ /^------/ or $_ =~ /^Model/);
						my $MTline = $_;

						if ($MTline =~ /^\(Output sorted by/)	# Set optimal model at first instance of output
							{
							$outputCount++;
								last if ($outputCount > 1);	# Only read in from first set of results; remaining sets are identical except for sort order
							}
						elsif ($MTline =~ /^Minimum/)	# Get optimal models
							{
							my ($criterion, $tempModel) = split(/:/, $MTline);
								$criterion =~ s/Minimum//;
								$criterion =~ s/model//;
								$criterion =~ s/\s+//g;
								$tempModel =~ s/\s+//g;
								
							$optimalModel{$criterion} = $tempModel;
							}
						elsif ($MTline =~ /Consider the use of AICc/ and $defaultCriterion eq "AIC")	# Reset optimality criterion if necessary
							{
							$MTcriterion = "AICc";
							}
						else	# Read in model and associated likelihood scores
							{
							my @MrAICline = split(/\s+/, $MTline);
							next unless ($MrAICline[0] eq $optimalModel{$MTcriterion});
							$MTmodel = $MrAICline[0];
								$MTmodel =~ s/IG$/+I+G/;
								$MTmodel =~ s/G$/+G/;
								$MTmodel =~ s/I$/+I/;
								$MTmodel .= " *" if ($MTcriterion ne $defaultCriterion);
							$nonclockL = $MrAICline[2];
							}
						}
				close RESULTS;
			unlink ("autoMT_phyml.txt");
			}
		
		# Print results to tab-delimited text file
			print "\tPrinting results to file autoMTresults.txt ...\n";
			print OUT "$file\t$MTmodel\t$nonclockL";
				printf OUT "\t$clockL\t$ntax\t$df\t$chiSquared\t$clockP" if ($clockTest);
			if ($testProgram eq "paup")
				{
				$LsetCommand =~ s/0\.0000/0\.0001/g;	# Remove impossible rate-matrix parameters
				print OUT "\t$LsetCommand";
				}
			print OUT "\n";
		
		my $fileTime = time - $fileZero;
			if ($fileTime > 3600)
				{
				printf "\n\tTotal time taken: %.1f hours\n\n", $fileTime / 3600;
				}
			else
				{
				print "\n\tTotal time taken: $fileTime seconds\n\n";
				}
		}

print OUT "\nNote: An asterisk following the chosen model means that the chosen criterion was overruled based on the recommendations of MrAIC.pl.\n" if ($testProgram eq "MrAIC");

close OUT;
exit(0);

# Subroutines used in the script

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks.\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "\tDOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "\tMac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "\tUnix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}
	
sub _subchisqrprob # Calculates one-tailed probability from chi-squared distribution; taken from the CPAN Perl module Statistics::Distributions by Michael Kospach
	{
	my ($n,$x) = @_;
	my $p;

	if ($x <= 0)
		{
		$p = 1;
		}
	elsif ($n > 100)
		{
		$p = _subuprob((($x / $n) ** (1/3) - (1 - 2/9/$n)) / sqrt(2/9/$n));
		}
	elsif ($x > 400)
		{
		$p = 0;
		}
	else
		{   
		my ($a, $i, $i1);
		if (($n % 2) != 0)
			{
			$p = 2 * _subuprob(sqrt($x));
			$a = sqrt(2/PI) * exp(-$x/2) / sqrt($x);
			$i1 = 1;
			}
		else
			{
			$p = $a = exp(-$x/2);
			$i1 = 2;
			}

		for ($i = $i1; $i <= ($n-2); $i += 2)
			{
			$a *= $x / $i;
			$p += $a;
			}
		}
	return $p;
}

sub _subuprob	# Also taken from the CPAN Perl module Statistics::Distributions by Michael Kospach
	{
	my ($x) = @_;
	my $p = 0; # if ($absx > 100)
	my $absx = abs($x);

	if ($absx < 1.9)
		{
		$p = (1 +
			$absx * (.049867347
			  + $absx * (.0211410061
			  	+ $absx * (.0032776263
				  + $absx * (.0000380036
					+ $absx * (.0000488906
					  + $absx * .000005383)))))) ** -16/2;
		}
	elsif ($absx <= 100)
		{
		for (my $i = 18; $i >= 1; $i--)
			{
			$p = $i / ($absx + $p);
			}
		$p = exp(-.5 * $absx * $absx) 
			/ sqrt(2 * PI) / ($absx + $p);
		}

	$p = 1 - $p if ($x<0);
	return $p;
}

sub printMT	# Prints ModelTEST block as appropriate
	{
	my $MTfile = shift;
	open (MT, ">$MTfile") or die "Cannot write to ModelTEST instruction file, $MTfile\n";
	
#	if ($refTree ne "NJ")
#		{
#		print MT "\nbegin paup;\n\tset warnroot = no;\nend;\n";
#		print MT "\nbegin trees;\n";
#		print MT "\t$refTree;\n";
#		print MT "end;\n";
#		}
	
	print MT "\n[! ***** MODELFIT BLOCK -- MODELTEST 3.4 *****]\n";
	print MT "\nBEGIN PAUP;\n";
	print MT "\tlog file = $modelFitLog replace;\n";
	
		# Print out instructions for reference tree
			if ($refTree eq "NJ") # Use default NJ tree
				{
				print MT "\n\t[The following command will calculate a NJ tree using the JC69 model of evolution]\n";
				print MT "\n\tDSet distance=JC objective=ME base=equal rates=equal pinv=0 subst=all negbrlen=setzero;\n";
				print MT "\tNJ showtree=no breakties=random;\n";
				}

	print MT "\n\t[! ***** BEGIN TESTING 56 MODELS OF EVOLUTION ***** ]\n";
	print MT "\n\tDefault lscores longfmt=yes;\n";
	print MT "\tSet criterion=like;\n";
	print MT "\n\t[! ** Model 1 of 56 * Calculating JC **]\n";
	print MT "\tlscores  1/ nst=1  base=equal  rates=equal  pinv=0\n";
	print MT "\tscorefile=$modelScores replace;\n";
	print MT "\t[! ** Model 2 of 56 * Calculating JC+I **]\n";
	print MT "\tlscores  1/ nst=1  base=equal  rates=equal  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 3 of 56 * Calculating JC+G **]\n";
	print MT "\tlscores  1/ nst=1  base=equal  rates=gamma  shape=est  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 4 of 56 * Calculating JC+I+G **]\n";
	print MT "\tlscores  1/ nst=1  base=equal  rates=gamma  shape=est  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 5 of 56 * Calculating F81 **]\n";
	print MT "\tlscores  1/ nst=1  base=est  rates=equal  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 6 of 56 * Calculating F81+I **]\n";
	print MT "\tlscores  1/ nst=1  base=est  rates=equal  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 7 of 56 * Calculating F81+G **]\n";
	print MT "\tlscores  1/ nst=1  base=est  rates=gamma  shape=est  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 8 of 56 * Calculating F81+I+G **]\n";
	print MT "\tlscores  1/ nst=1  base=est  rates=gamma  shape=est  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 9 of 56 * Calculating K80 **]\n";
	print MT "\tlscores  1/ nst=2  base=equal  tratio=est  rates=equal  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 10 of 56 * Calculating K80+I **]\n";
	print MT "\tlscores  1/ nst=2  base=equal  tratio=est  rates=equal  pin=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 11 of 56 * Calculating K80+G **]\n";
	print MT "\tlscores  1/ nst=2  base=equal  tratio=est  rates=gamma  shape=est  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 12 of 56 * Calculating K80+I+G **]\n";
	print MT "\tlscores  1/ nst=2  base=equal  tratio=est  rates=gamma  shape=est pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 13 of 56 * Calculating HKY **]\n";
	print MT "\tlscores  1/ nst=2  base=est  tratio=est  rates=equal  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 14 of 56 * Calculating HKY+I **]\n";
	print MT "\tlscores  1/ nst=2  base=est  tratio=est  rates=equal  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 15 of 56 * Calculating HKY+G **]\n";
	print MT "\tlscores  1/ nst=2  base=est  tratio=est  rates=gamma  shape=est  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 16 of 56 * Calculating HKY+I+G **]\n";
	print MT "\tlscores  1/ nst=2  base=est  tratio=est  rates=gamma  shape=est  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 17 of 56 * Calculating TrNef **]                            [a b c d e f]\n";
	print MT "\tlscores  1/ nst=6  base=equal  rmat=est  rclass=(a b a a e a)  rates=equal  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 18 of 56 * Calculating TrNef+I **]\n";
	print MT "\tlscores  1/ nst=6  base=equal  rmat=est  rates=equal  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 19 of 56 * Calculating TrNef+G **]\n";
	print MT "\tlscores  1/ nst=6  base=equal rmat=est  rates=gamma  shape=est  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 20 of 56 * Calculating TrNef+I+G **]\n";
	print MT "\tlscores  1/ nst=6  base=equal  rmat=est rates=gamma  shape=est  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 21 of 56 * Calculating TrN **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=equal  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 22 of 56 * Calculating TrN+I **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=equal  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 23 of 56 * Calculating TrN+G **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=gamma  shape=est  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 24 of 56 * Calculating TrN+I+G **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=gamma  shape=est  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 25 of 56 * Calculating K3P **]                              [a b c d e f]\n";
	print MT "\tlscores  1/ nst=6  base=equal  rmat=est  rclass=(a b c c b a)  rates=equal  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 26 of 56 * Calculating K3P+I **]\n";
	print MT "\tlscores  1/ nst=6  base=equal  rmat=est  rates=equal  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 27 of 56 * Calculating K3P+G **]\n";
	print MT "\tlscores  1/ nst=6  base=equal rmat=est  rates=gamma  shape=est  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 28 of 56 * Calculating K3P+I+G **]\n";
	print MT "\tlscores  1/ nst=6  base=equal  rmat=est rates=gamma  shape=est  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 29 of 56 * Calculating K3Puf **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=equal  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 30 of 56 * Calculating K3Puf+I **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=equal  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 31 of 56 * Calculating K3Puf+G **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=gamma  shape=est  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 32 of 56 * Calculating K3Puf+I+G **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=gamma  shape=est  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 33 of 56 * Calculating TIMef **]                            [a b c d e f]\n";
	print MT "\tlscores  1/ nst=6  base=equal  rmat=est  rclass=(a b c c e a)  rates=equal  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 34 of 56 * Calculating TIMef+I **]\n";
	print MT "\tlscores  1/ nst=6  base=equal  rmat=est  rates=equal  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 35 of 56 * Calculating TIMef+G **]\n";
	print MT "\tlscores  1/ nst=6  base=equal rmat=est  rates=gamma  shape=est  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 36 of 56 * Calculating TIMef+I+G **]\n";
	print MT "\tlscores  1/ nst=6  base=equal  rmat=est rates=gamma  shape=est  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 37 of 56 * Calculating TIM **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=equal  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 38 of 56 * Calculating TIM+I **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=equal  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 39 of 56 * Calculating TIM+G **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=gamma  shape=est  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 40 of 56 * Calculating TIM+I+G **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=gamma  shape=est  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 41 of 56 * Calculating TVMef **]                            [a b c d e f]\n";
	print MT "\tlscores  1/ nst=6  base=equal  rmat=est  rclass=(a b c d b e)  rates=equal  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 42 of 56 * Calculating TVMef+I **]\n";
	print MT "\tlscores  1/ nst=6  base=equal  rmat=est  rates=equal  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 43 of 56 * Calculating TVMef+G **]\n";
	print MT "\tlscores  1/ nst=6  base=equal rmat=est  rates=gamma  shape=est  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 44 of 56 * Calculating TVMef+I+G **]\n";
	print MT "\tlscores  1/ nst=6  base=equal  rmat=est rates=gamma  shape=est  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 45 of 56 * Calculating TVM **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=equal  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 46 of 56 * Calculating TVM+I **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=equal  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 47 of 56 * Calculating TVM+G **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=gamma  shape=est  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 48 of 56 * Calculating TVM+I+G **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=gamma  shape=est  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 49 of 56 * Calculating SYM **]                               [a b c d e f]\n";
	print MT "\tlscores  1/ nst=6  base=equal  rmat=est  rclass= (a b c d e f)  rates=equal  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 50 of 56 * Calculating SYM+I **]\n";
	print MT "\tlscores  1/ nst=6  base=equal  rmat=est  rates=equal  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 51 of 56 * Calculating SYM+G **]\n";
	print MT "\tlscores  1/ nst=6  base=equal  rmat=est  rates=gamma  shape=est pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 52 of 56 * Calculating SYM+I+G **]\n";
	print MT "\tlscores  1/ nst=6  base=equal  rmat=est  rates=gamma  shape=est pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 53 of 56 * Calculating GTR **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=equal  pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 54 of 56 * Calculating GTR+I **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=equal  pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 55 of 56 * Calculating GTR+G **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=gamma  shape=est pinv=0\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\t[! ** Model 56 of 56 * Calculating GTR+I+G **]\n";
	print MT "\tlscores  1/ nst=6  base=est  rmat=est  rates=gamma  shape=est pinv=est\n";
	print MT "\tscorefile=$modelScores  append;\n";
	print MT "\n\t[! NOTE:The molecular clock hypothesis is not included in this PAUP block for reasons of\n";
	print MT "\tcomputing time. This hypothesis may be tested once we have chosen one of the above\n";
	print MT "\tmodels, simply calculating the log likelihood score of the chosen model with the\n";
	print MT "\tmolecular clock enforced and comparing it with the log likelihood previously obtained\n";
	print MT "\twithout enforcing the molecular clock. In this case, the molecular clock is the\n";
	print MT "\tnull hypothesis. The number of degrees of freedom is the number of OTUs - 2]\n";
	print MT "\n\tlog stop;\n";
	print MT "end;\n";
	print MT "\n[\nlscores: tree/s\n";
	print MT "nst: number of substitution types\n";
	print MT "rmat: rate matrix\n";
	print MT "base= nucleotide frequencies\n";
	print MT "rates= rate of evolution for variable sites (same\n";
	print MT "shape= alpha parameter of the gamma distribution\n";
	print MT "pinv= proportion of invariable sites\n]\n";
	print MT "\n[The likelihood ratio test statistic is calculated as twice the difference between\n";
	print MT "the log likelihood scores of the two models constrasted. When the model representing\n";
	print MT "the null hypothesis is a special case of the alternate model, this statistic\n";
	print MT "fits a chi-square distribution with a number of degrees of freedom equal to the\n";
	print MT "number of parameters that freely vary between the two models. A detailed explanation\n";
	print MT "of the test of the model of DNA substitution can be found in Huelsenbeck and Crandall\n";
	print MT "Annu. Rev. Ecol. Syst. 1997. 28: 437-466]\n";
	print MT "\n[!***  END OF MODELTEST BLOCK  ***]\n";

	# Add terminal quit statement
		print MT "quit warntsave=no;\n";

	close MT;		
	return;
	}

# Version history
#
# 	v1.0 (March 2, 2006)
#		- initial release
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is available at
# http://www.gnu.org/copyleft/gpl.html or by writing to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Fifth Floor, Boston, MA, 02110-1301, USA.
